import numpy as np
import scipy.stats as stats
from scipy.interpolate import PchipInterpolator
import matplotlib.pyplot as plt


class Histogram:
    def __init__(self, sample, bin_size, confidence_level=0.95):
        self.sample = sample
        self.bin_size = bin_size
        self.confidence_level = confidence_level
        self.bins = int((np.amax(sample) - np.amin(sample)) // bin_size + 1)
        self.__min = np.amin(sample) // bin_size - 1
        self.__shift = self.__min * bin_size + bin_size / 2
        self.__shifted_sample = sample - self.__shift
        self._x, self._y = self.__bin(self.__shifted_sample, bin_size, self.bins)

    def __bin(self, sample, bin_size, bins):
        assert bin_size > 0
        assert bins > 0
        assert type(bins) == int
        assert type(bin_size) in [int, float]

        x = np.linspace(0, bins * bin_size, bins + 1, endpoint=True)
        y = np.zeros(bins + 1)

        for s in sample:
            index = int((s // bin_size))
            if index >= bins:  # Ensure index is within bounds
                index = bins - 1
            y[index] += 1

        return x, y

    @property
    def x(self):
        return self._x + self.__shift + self.bin_size / 2

    @property
    def y(self):
        return self._y / (np.sum(self._y) * self.bin_size)
    
    @property
    def e(self):
        return self.error_estimate(self._y) / (np.sum(self._y) * self.bin_size)
                                       
    @property
    def counts(self):
        return self._y

    @property
    def counts_error(self):
        return self.error_estimate(self._y)
    
    def poisson_confidence_interval(self, k):
        alpha = 1 - self.confidence_level
        lower_bound = 0.5 * stats.chi2.ppf(alpha / 2, 2 * k)
        upper_bound = 0.5 * stats.chi2.ppf(1 - alpha / 2, 2 * (k + 1))
        
        if type(k) == np.ndarray:
            lower_bound[k == 0] = 0
        return lower_bound, upper_bound
    
    def error_estimate(self, k):
        lower_bound, upper_bound = self.poisson_confidence_interval(k)
        return (upper_bound - lower_bound) / 2
    
    def FWHM(self, resolution):
        y = PchipInterpolator(self.x, self.y)

        max_value = np.amax(self.y)
        half_max = max_value / 2

        peak_x = self.x[np.argmax(self.y)]

        min_x = np.amin(self.x)
        max_x = np.amax(self.x)

        left = bisection(self.x, y, half_max, resolution, min_x, peak_x)
        right = bisection(self.x, y, half_max, resolution, peak_x, max_x)

        return np.abs(right - left)
    
    def plot(self, **kwargs):
        
        x = self.x if 'x' not in kwargs else kwargs.pop('x')
        
        plt.rcParams["font.family"] = "monospace"
        plt.errorbar(x=self.x, 
                     y=self.y, 
                     yerr=self.e, 
                     fmt="s", 
                     color='black' if "color" not in kwargs else kwargs.pop("color"), 
                     zorder=1 if "zorder" not in kwargs else kwargs.pop("zorder"),
                     markerfacecolor='none' if "markerfacecolor" not in kwargs else kwargs.pop("markerfacecolor"),
                     label="Data" if "label" not in kwargs else kwargs.pop("label"))

    def show(self):
        plt.legend()
        plt.show()
        
def bisection(x, y, value, tolerance, left, right):
    assert len(x) > 1
    assert tolerance > 0

    if right - left <= tolerance:
        return (left + right) / 2

    mid = (left + right) / 2

    if y(mid) < value < y(right) or y(mid) > value > y(right):
        return bisection(x, y, value, tolerance, mid, right)
    else:
        return bisection(x, y, value, tolerance, left, mid)


if __name__ == '__main__':
    sample = np.random.normal(0, 1, 10000000)
    hist = Histogram(sample, 0.1)
    hist.plot(label='Data 1')
    hist.show()
    print(hist.FWHM(1e-5))


import numpy as np
import matplotlib.pyplot as plt
from histogram import Histogram
from iminuit import Minuit
from iminuit.cost import LeastSquares


def decay(x, tau):
    return tau * (1 - np.exp(-x / tau))


def model(x, dt, risetime, decaytime, lightyield):
    ctr = np.exp(
        - (lightyield / (decaytime - risetime)) * (decay(np.abs(x + dt), decaytime) - decay(np.abs(x + dt), risetime)))
    area = np.trapz(ctr, x)

    return ctr / area


class CTR(Histogram):
    def __init__(self, sample, bin_size, confidence_level=0.95):
        super().__init__(sample, bin_size, confidence_level)
        self.sample_size = len(self.sample)

    def fit(self, risetime, decaytime, lightyield):
        min_fval = None
        m_min = None
        for _ in range(1000):
            lsq = LeastSquares(self.x, self.y, self.e, model)
            m = Minuit(lsq,
                       dt=np.random.normal(np.mean(self.sample), np.std(self.sample)),
                       risetime=risetime,
                       decaytime=decaytime,
                       lightyield=lightyield * np.random.uniform(0.5, 1.5))

            m.fixed['risetime'] = True
            m.fixed['decaytime'] = True

            m.migrad()
            m.hesse()

            if min_fval is None or m.fval < min_fval:
                min_fval = m.fval
                m_min = m

        m = m_min

        m.fixed['risetime'] = False
        m.fixed['decaytime'] = True
        m.fixed['lightyield'] = True

        m.migrad()
        m.hesse()

        m.fixed['risetime'] = True
        m.fixed['decaytime'] = False
        m.fixed['lightyield'] = True

        m.migrad()
        m.hesse()

        m.fixed['risetime'] = True
        m.fixed['decaytime'] = True
        m.fixed['lightyield'] = False

        m.migrad()
        m.hesse()


        self.dt = m.values['dt']
        self.risetime = m.values['risetime']
        self.decaytime = m.values['decaytime']
        self.lightyield = m.values['lightyield']
        self.dt_error = m.errors['dt']
        self.risetime_error = m.errors['risetime']
        self.decaytime_error = m.errors['decaytime']
        self.lightyield_error = m.errors['lightyield']
        self.chi2 = m.fval / (len(self.x) - len(m.values))
        self.ndof = len(self.x) - len(m.values)

    def plot(self):
        plt.rcParams["font.family"] = "monospace"
        plt.rcParams["font.sans-serif"] = ["DejaVu Sans", "Arial", "Liberation Sans"]

        fit_info = [
            f'$\chi^2$\t      ${self.chi2:.3f}$\n',
            f'$\Delta t$ \t\t${self.dt:.2E} \pm {self.dt_error:.1E}$',
            f'Rise Time \t${self.risetime:.2E} \pm {self.risetime_error:.1E}$',
            f'Decay Time \t${self.decaytime:.2E} \pm {self.decaytime_error:.1E}$',
            f'Light Yield \t${self.lightyield:.2E} \pm {self.lightyield_error:.1E}$\n',
            f'FWHM \t\t${self.FWHM(10E-12):.2E}$',
            f'Sample Size \t${self.sample_size}$\n',
        ]

        x = np.linspace(-np.max(self.x), np.max(self.x), 1000)
        plt.errorbar(self.x, self.y, yerr=self.e, fmt='o', label='Data', zorder=1, color='black')
        plt.plot(x, model(x, self.dt, self.risetime, self.decaytime, self.lightyield), label='Fit', zorder=2, color='black', linestyle='--',
                 linewidth=2)

        plt.legend(title='\n'.join(fit_info))
        plt.xlabel('Time Difference (s)')
        plt.ylabel('Probability Density')
        plt.grid(True)
        






import numpy as np
import matplotlib.pyplot as plt 
from .histogram import Histogram
from iminuit import Minuit
from iminuit.cost import LeastSquares


def model(x, mu, sigma):
    return 1/(sigma * np.sqrt(2 * np.pi)) * np.exp(-0.5 * ((x - mu) / sigma) ** 2)


class Gaussian(Histogram):
    def __init__(self, sample, bin_size, confidence_level=0.95):
        super().__init__(sample, bin_size, confidence_level)
        self.sample_size = len(self.sample)
        
    def fit(self):
        min_fval = None
        m_min = None
        for _ in range(1000):
            lsq = LeastSquares(self.x, self.y, self.e, model)
            m = Minuit(lsq, mu=np.random.normal(1, 0.1) * np.mean(self.sample), sigma=np.random.normal(1, 0.1) * np.std(self.sample))
            
            m.migrad()
            m.hesse()

            if min_fval is None or m.fval < min_fval:
                min_fval = m.fval
                m_min = m
        m = m_min
        
        self.mu = m.values['mu']
        self.sigma = m.values['sigma']
        self.mu_error = m.errors['mu']
        self.sigma_error = m.errors['sigma']
        self.chi2 = m.fval / (len(self.x) - len(m.values))
        self.ndof = len(self.x) - len(m.values)
        
        
    def plot(self):
        plt.rcParams["font.family"] = "monospace"
        fit_info = [
            f'$\chi^2$\t      ${self.chi2:.3f}$',
            f'Mean \t\t${self.mu:.3E} \pm {self.mu_error:.2E}$',
            f'Std. Dev. \t${self.sigma:.3E} \pm {self.sigma_error:.2E}$'
        ]
        
        plt.errorbar(self.x, self.y, yerr=self.e, fmt='o', label='Data', zorder=1, color='black')
        plt.plot(self.x, model(self.x, self.mu, self.sigma), label='Fit', zorder=2, color='black', linestyle='--', linewidth=2)
        
        plt.legend(title='\n'.join(fit_info))
        plt.show()
        
   

